import os
import subprocess
import numpy as np
from soundfile import SoundFile
import sox
from dataclasses import dataclass, field

from sox.file_info import channels, sample_rate


def file_ext(filepath: str) -> str:
    """Gets the extension of a file."""
    _, ext = os.path.splitext(filepath)
    return ext.split(".")[1]


def _subtype_to_bd(sub):
    sub_map = {
        "PCM_8": 8,
        "PCM_16": 16,
        "PCM_24": 24,
        "PCM_32": 32,
    }
    if sub not in sub_map:
        return None
    else:
        return sub_map[sub]


def _bd_to_subtype(bd: int) -> str:
    if bd in [8, 16, 24, 32]:
        return f"PCM_{bd}"
    else:
        return None


@dataclass
class AudioConfig:
    sample_rate: float
    data_width: int
    channels: int

    @property
    def sub_type(self) -> str:
        return _bd_to_subtype(self.data_width)

    @sub_type.setter
    def sub_type(self, st: str):
        self.data_width = _subtype_to_bd(st)


@dataclass
class AudioData:
    cfg: AudioConfig
    data: np.ndarray = field(default=None)
    time: np.ndarray = field(init=False, default=None)

    def __post_init__(self):
        if self.data is not None:
            self.time = np.linspace(
                0.0, self.data.shape[0] / self.cfg.sample_rate, self.data.shape[0]
            )

    def __len__(self) -> int:
        return int(self.data.shape[0])

    # direct_fmts = [".wav", ".aiff", ".flac", ".ogg"]
    # convertible_fmts = direct_fmts + [".mp3"]
    # valid_modes = ["r", "w", "c"]


@dataclass
class AudioFile:
    file: str
    mode: str

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        """
        Leaving context manager. Make sure any open files are closed.
        """
        pass

    @classmethod
    def file_info(cls, file: str) -> AudioConfig:
        sample_rate = int(sox.file_info.sample_rate(file))
        data_width = sox.file_info.bitdepth(file)
        channels = sox.file_info.channels(file)
        return AudioConfig(sample_rate, data_width, channels)

    def read(self) -> AudioData:
        """
        Reads in audio file into an AudioData instance and returns the obj
        """
        if "r" not in self.mode:
            raise Exception("read mode is not enabled.")
        if not os.path.exists(self.file):
            raise FileNotFoundError("Audio source file not found.")

        # Read in audio data using SoundFile.
        with SoundFile(self.file, "r") as sfo:
            cfg = AudioFile.file_info(self.file)
            audio = sfo.read(dtype="int32")
            ad = AudioData(cfg, audio)
            # 24 bit data gets read in as 32 bit so we need to fix it.
            if ad.cfg.data_width == 24:
                ad.data = ad.data >> 8

        return ad

    def convert(self, ofile, cfg: AudioConfig, overwrite=True) -> None:
        """
        Convert the audio file to the requested output.
        The source file gets updated to point to the converted file.
        """
        if "c" not in self.mode:
            raise Exception("Conversion not allowed in this context.")

        if not os.path.exists(self.file):
            raise FileNotFoundError("Audio source file not found.")

        if os.path.exists(ofile) and not overwrite:
            raise FileExistsError(
                "Conversion output file already exists and overwrite is disabled."
            )

        ofile_ext = file_ext(ofile)
        # if ofile_ext not in AudioFile.convertible_fmts:
        #     raise Exception("Output file is not a convertible audio file type.")

        # Default conversion command using SoX.
        tfm = sox.Transformer()
        tfm.set_output_format(
            ofile_ext, cfg.sample_rate, cfg.data_width, channels=cfg.channels
        )
        tfm.build_file(self.file, ofile)

    def write(self, ad: AudioData):
        """
        Save/reconvert input file
        """
        if "w" not in self.mode:
            raise Exception("Writing is not allowed in this context.")

        # Write file using SoundFile.
        with SoundFile(
            self.file, "w", ad.cfg.sample_rate, ad.cfg.channels, ad.cfg.sub_type
        ) as sfo:
            # Convert 24 bit to 32 bit before writing.
            if ad.cfg.data_width == 24:
                cnv_data = ad.data << 8
            else:
                cnv_data = ad.data

            sfo.write(cnv_data)

    def play(self, duration: int = 0):
        """
        Playback an audiofile using aplay. Must be a direct format.
        This can be called in any context and will check for valid types in case it's a post-write play.
        If you just want to play a file, open it in the read context.
        """
        if not os.path.exists(self.file):
            raise Exception("Playback file does not exist.")

        ext = file_ext(self.file)

        # if ext not in AudioFile.direct_fmts:
        #     raise Exception(f"Cannot play {ext} file types. Please convert first.")

        play_cmd = ["aplay", "-d", str(duration), self.file]
        try:
            subprocess.check_call(play_cmd)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Error playing {self.file}")
