import argparse
from io import TextIOWrapper
import sys
import os

from . import templates as abies_templates
import importlib.resources
import yaml

import abies.simulator as sim

"""
commands can be used directly, or using a config file with options.

Commands:

# Generate a simulation model 
abies gen_model --source <source.v> --build <build_dir>
# generate all simulation models
abies gen_model --config <project.yaml>
"""


def read_abies_template(name: str) -> str:
    """"""
    with importlib.resources.open_text("abies.templates", name) as fp:
        """"""
        content = fp.read()

    return content


def read_project_file(project_file: str):
    """"""
    with open(project_file, "r") as fp:
        return yaml.safe_load(fp.read())


def read_yaml(fp: TextIOWrapper):
    return yaml.safe_load(fp.read())


def gen_model_cb(args):
    """"""
    print(args)
    if args.project is not None:
        with open(args.project, "r") as fp:
            prj = read_yaml(fp)
    else:
        with importlib.resources.open_text(
            abies_templates, "default_project.yaml"
        ) as fp:
            prj = read_yaml(fp)

    # Override options
    if args.source is not None:
        prj["modules"].append(args.source)

    if args.build is not None:
        prj["build_dir"] = args.build

    if args.rtl:
        prj["rtl_dir"] = args.rtl

    for module in prj["modules"]:
        sim.build_verilator_model(module, prj["build_dir"])


def add_common_args(parser: argparse.ArgumentParser):
    """"""
    parser.add_argument(
        "-p", "--project", metavar="<config.yaml>", action="store", help="project file."
    )
    parser.add_argument(
        "-b", "--build", metavar="<build_dir>", action="store", help="build directory."
    )
    parser.add_argument(
        "-r", "--rtl", metavar="<rtl_dir>", action="store", help="rtl directory."
    )


def create_parser():
    """"""
    parser = argparse.ArgumentParser(description="Abies Audio Processing Framework")

    subparsers = parser.add_subparsers(help="sp help")

    gen_model_sp = subparsers.add_parser("gen_model", help="gen help")
    add_common_args(gen_model_sp)
    gen_model_sp.add_argument(
        "-s",
        "--source",
        metavar="<source.v>",
        action="store",
        help="verilog source file.",
    )
    gen_model_sp.set_defaults(func=gen_model_cb)
    return parser


def main():

    parser = create_parser()

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    args.func(args)
