from ctypes import CDLL, c_void_p, c_int, c_char_p
from .model import Model

import numpy as np
from typing import Protocol
from abc import abstractmethod
from pathlib import Path

c_model_api = {
    "create_model": ([c_char_p], c_void_p),
    "destroy_model": ([c_void_p], None),
    "reset_model": ([c_void_p, c_int], None),
    "process_model": (
        [
            c_void_p,
            np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags="C_CONTIGUOUS"),
            c_int,
            np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags="C_CONTIGUOUS"),
            c_int,
            c_int,
        ],
        c_int,
    ),
}


def load_api(lib: CDLL, api_def: dict[str, tuple]):
    """"""
    for fn_name, arg_types in api_def.items():
        func = getattr(lib, fn_name)
        func.argtypes = arg_types[0]
        func.restype = arg_types[1]
        # setattr(lib, fn_name, func)  # Do I need to do this?


class CModel(Model):
    def __init__(
        self, lib_path: Path, trace: str = "", model_api: dict[str, tuple] = c_model_api
    ) -> None:
        """"""
        self._lib = CDLL(str(lib_path))
        load_api(self._lib, model_api)
        self._m = self._lib.create_model(trace.encode("UTF-8"))

    def __del__(self):
        self._lib.destroy_model(self._m)

    def reset(self, cycles: c_int) -> None:
        self._lib.reset_model(self._m, cycles)

    def process(self, idata: np.ndarray, clk_div: int = 10) -> np.ndarray:
        """"""
        odata = np.zeros(len(idata), dtype=np.int32)
        error = self._lib.process_model(
            self._m, idata, len(idata), odata, len(odata), clk_div
        )
        return odata
